-module(erl_parse).
-define(THIS_MODULE, erl_parse).
-export([parse/1, parse_and_scan/1, format_error/1]).

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

-export([parse_form/1,parse_exprs/1,parse_term/1]).
-export([normalise/1,abstract/1,tokens/1,tokens/2]).
-export([abstract/2]).
-export([inop_prec/1,preop_prec/1,func_prec/0,max_prec/0]).

-compile({inline,[{mkop,3},{mkop,2},{line,1}]}).	%Thrilling!

%% mkop(Op, Arg) -> {op,Line,Op,Arg}.
%% mkop(Left, Op, Right) -> {op,Line,Op,Left,Right}.

mkop(L, {Op,Pos}, R) -> {op,Pos,Op,L,R}.

mkop({Op,Pos}, A) -> {op,Pos,Op,A}.

%% keep track of line info in tokens
line(Tup) -> element(2, Tup).

%% Entry points compatible to old erl_parse.
%% These really suck and are only here until Calle gets multiple
%% entry points working.

parse_form(Tokens) ->
    parse(Tokens).

parse_exprs(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],Exprs}]}} ->
	    {ok,Exprs};
	{error,E} -> {error,E}
    end.

parse_term(Tokens) ->
    case parse([{atom,0,f},{'(',0},{')',0},{'->',0}|Tokens]) of
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[Expr]}]}} ->
	    case catch normalise(Expr) of
		{'EXIT',R} ->
		    {error,{line(Expr),?MODULE,"bad term"}};
		Term -> {ok,Term}
	    end;
	{ok,{function,Lf,f,0,[{clause,Lc,[],[],[E1,E2|Es]}]}} ->
	    {error,{line(E2),?MODULE,"bad term"}};
	{error,E} -> {error,E}
    end.

%% build_attribute(AttrName, AttrValue) ->
%%	{attribute,Line,module,Module}
%%	{attribute,Line,export,Exports}
%%	{attribute,Line,import,Imports}
%%	{attribute,Line,record,{Name,Inits}}
%%	{attribute,Line,file,{Name,Line}}
%%	{attribute,Line,Name,Val}

build_attribute({atom,La,module}, Val) ->
    case Val of
	[{atom,Lm,Module}] -> {attribute,La,module,Module};
	Other -> return_error(La, "bad module declaration")
    end;
build_attribute({atom,La,export}, Val) ->
    case Val of
	[ExpList] ->
	    {attribute,La,export,farity_list(ExpList)};
	Other -> return_error(La, "bad export declaration")
    end;
build_attribute({atom,La,import}, Val) ->
    case Val of
	[{atom,Lm,Mod},ImpList] ->
	    {attribute,La,import,{Mod,farity_list(ImpList)}};
	Other -> return_error(La, "bad import declaration")
    end;
build_attribute({atom,La,record}, Val) ->
    case Val of
	[{atom,Ln,Record},RecTuple] ->
	    {attribute,La,record,{Record,record_tuple(RecTuple)}};
	Other -> return_error(La, "bad record declaration")
    end;
build_attribute({atom,La,file}, Val) ->
    case Val of
	[{string,Ln,Name},{integer,Ll,Line}] ->
	    {attribute,La,file,{Name,Line}};
	Other -> return_error(La, "bad file declaration")
    end;
build_attribute({atom,La,Attr}, Val) ->
    case Val of
	[Expr] ->
	    {attribute,La,Attr,term(Expr)};
	Other -> return_error(La, "bad attribute")
    end.

farity_list({cons,Lc,{op,Lo,'/',{atom,La,A},{integer,Li,I}},Tail}) ->
    [{A,I}|farity_list(Tail)];
farity_list({nil,Ln}) -> [];
farity_list(Other) ->
    return_error(line(Other), "bad function arity").

record_tuple({tuple,Lt,Fields}) ->
    record_fields(Fields);
record_tuple(Other) ->
    return_error(line(Other), "bad record declaration").

record_fields([{atom,La,A}|Fields]) ->
    [{record_field,La,{atom,La,A}}|record_fields(Fields)];
record_fields([{match,Lm,{atom,La,A},Expr}|Fields]) ->
    [{record_field,La,{atom,La,A},Expr}|record_fields(Fields)];
record_fields([Other|Fields]) ->
    return_error(line(Other), "bad record field");
record_fields([]) -> [].

term(Expr) ->
    case catch normalise(Expr) of
	{'EXIT',R} -> return_error(line(Expr), "bad attribute");
	Term -> Term
    end.

%% build_function([Clause]) -> {function,Line,Name,Arity,[Clause]}

build_function(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {function,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_rule([Clause]) -> {rule,Line,Name,Arity,[Clause]'}

build_rule(Cs) ->
    Name = element(3, hd(Cs)),
    Arity = length(element(4, hd(Cs))),
    {rule,line(hd(Cs)),Name,Arity,check_clauses(Cs, Name, Arity)}.

%% build_fun(Line, [Clause]) -> {'fun',Line,{clauses,[Clause]}}.

build_fun(Line, Cs) ->
    Arity = length(element(4, hd(Cs))),
    {'fun',Line,{clauses,check_clauses(Cs, 'fun', Arity)}}.

check_clauses(Cs, Name, Arity) ->
     mapl(fun ({clause,L,N,As,G,B}) when N == Name, length(As) == Arity ->
		 {clause,L,As,G,B};
	     ({clause,L,N,As,G,B}) ->
		 return_error(L, "head mismatch") end, Cs).

%% mapl(F,List)
%% an alternative map which always maps from left to right
%% and makes it possible to interrupt the mapping with throw on
%% the first occurence from left as expected.
%% can be removed when the jam machine (and all other machines)
%% uses the standardized (Erlang 5.0) evaluation order (from left to right)
mapl(F, [H|T]) ->
	V = F(H),
	[V | mapl(F,T)];
mapl(_, []) ->
	[].

%% normalise(AbsTerm)
%% abstract(Term)
%%  Convert between the abstract form of a term and a term.

normalise({char,_,C}) -> C;
normalise({integer,_,I}) -> I;
normalise({float,_,F}) -> F;
normalise({atom,_,A}) -> A;
normalise({string,_,S}) -> S;
normalise({nil,_}) -> [];
normalise({bin,_,Fs}) ->
    {value, B, _} =
	eval_bits:expr_grp(Fs, [],
			   fun(E, _) ->
				   {value, normalise(E), []}
			   end, [], true),
    B;
normalise({cons,_,Head,Tail}) ->
    [normalise(Head)|normalise(Tail)];
normalise({tuple,_,Args}) ->
    list_to_tuple(normalise_list(Args));
%% Special case for unary +/-.
normalise({op,_,'+',{char,_,I}}) -> I;
normalise({op,_,'+',{integer,_,I}}) -> I;
normalise({op,_,'+',{float,_,F}}) -> F;
normalise({op,_,'-',{char,_,I}}) -> -I;		%Weird, but compatible!
normalise({op,_,'-',{integer,_,I}}) -> -I;
normalise({op,_,'-',{float,_,F}}) -> -F.

normalise_list([H|T]) ->
    [normalise(H)|normalise_list(T)];
normalise_list([]) ->
    [].

abstract(T) when integer(T) -> {integer,0,T};
abstract(T) when float(T) -> {float,0,T};
abstract(T) when atom(T) -> {atom,0,T};
abstract([]) -> {nil,0};
abstract(B) when binary(B) ->
    {bin, 0, lists:map(fun(Byte) ->
			       {bin_element, 0,
				{integer, 0, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T]) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C]);
abstract([H|T]) ->
    {cons,0,abstract(H),abstract(T)};
abstract(Tuple) when tuple(Tuple) ->
    {tuple,0,abstract_list(tuple_to_list(Tuple))}.

abstract_string([C|T], String) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String]);
abstract_string([], String) ->
    {string, 0, lists:reverse(String)};
abstract_string(T, String) ->
    not_string(String, abstract(T)).

not_string([C|T], Result) ->
    not_string(T, {cons, 0, {integer, 0, C}, Result});
not_string([], Result) ->
    Result.

abstract_list([H|T]) ->
    [abstract(H)|abstract_list(T)];
abstract_list([]) ->
    [].

%%% abstract/2 keeps the line number
abstract(T, Line) when integer(T) -> {integer,Line,T};
abstract(T, Line) when float(T) -> {float,Line,T};
abstract(T, Line) when atom(T) -> {atom,Line,T};
abstract([], Line) -> {nil,Line};
abstract(B, Line) when binary(B) ->
    {bin, Line, lists:map(fun(Byte) ->
			       {bin_element, Line,
				{integer, Line, Byte}, default, default}
		       end,
		       binary_to_list(B))};
abstract([C|T], Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C], Line);
abstract([H|T], Line) ->
    {cons,Line,abstract(H, Line),abstract(T, Line)};
abstract(Tuple, Line) when tuple(Tuple) ->
    {tuple,Line,abstract_list(tuple_to_list(Tuple), Line)}.

abstract_string([C|T], String, Line) when integer(C), 0 =< C, C < 256 ->
    abstract_string(T, [C|String], Line);
abstract_string([], String, Line) ->
    {string, Line, lists:reverse(String)};
abstract_string(T, String, Line) ->
    not_string(String, abstract(T, Line), Line).

not_string([C|T], Result, Line) ->
    not_string(T, {cons, Line, {integer, Line, C}, Result}, Line);
not_string([], Result, Line) ->
    Result.

abstract_list([H|T], Line) ->
    [abstract(H, Line)|abstract_list(T, Line)];
abstract_list([], Line) ->
    [].

%% tokens(AbsTerm) -> [Token]
%% tokens(AbsTerm, More) -> [Token]
%%  Generate a list of tokens representing the abstract term.

tokens(Abs) ->
    tokens(Abs, []).

tokens({char,L,C}, More) -> [{char,L,C}|More];
tokens({integer,L,N}, More) -> [{integer,L,N}|More];
tokens({float,L,F}, More) -> [{float,L,F}|More];
tokens({atom,L,A}, More) -> [{atom,L,A}|More];
tokens({var,L,V}, More) -> [{var,L,V}|More];
tokens({string,L,S}, More) -> [{string,L,S}|More];
tokens({nil,L}, More) -> [{'[',L},{']',L}|More];
tokens({cons,L,Head,Tail}, More) ->
    [{'[',L}|tokens(Head, tokens_tail(Tail, More))];
tokens({tuple,L,[]}, More) ->
    [{'{',L},{'}',L}|More];
tokens({tuple,L,[E|Es]}, More) ->
    [{'{',L}|tokens(E, tokens_tuple(Es, line(E), More))].

tokens_tail({cons,L,Head,Tail}, More) ->
    [{',',L}|tokens(Head, tokens_tail(Tail, More))];
tokens_tail({nil,L}, More) ->
    [{']',L}|More];
tokens_tail(Other, More) ->
    L = line(Other),
    [{'|',L}|tokens(Other, [{']',L}|More])].

tokens_tuple([E|Es], Line, More) ->
    [{',',Line}|tokens(E, tokens_tuple(Es, line(E), More))];
tokens_tuple([], Line, More) ->
    [{'}',Line}|More].

%% Give the relative precedences of operators.

inop_prec('=') -> {150,100,100};
inop_prec('!') -> {150,100,100};
inop_prec('orelse') -> {160,150,150};
inop_prec('andalso') -> {200,160,160};
inop_prec('==') -> {300,200,300};
inop_prec('/=') -> {300,200,300};
inop_prec('=<') -> {300,200,300};
inop_prec('<') -> {300,200,300};
inop_prec('>=') -> {300,200,300};
inop_prec('>') -> {300,200,300};
inop_prec('=:=') -> {300,200,300};
inop_prec('=/=') -> {300,200,300};
inop_prec('++') -> {300,300,400};
inop_prec('--') -> {300,300,400};
inop_prec('+') -> {400,400,500};
inop_prec('-') -> {400,400,500};
inop_prec('bor') -> {400,400,500};
inop_prec('bxor') -> {400,400,500};
inop_prec('bsl') -> {400,400,500};
inop_prec('bsr') -> {400,400,500};
inop_prec('or') -> {400,400,500};
inop_prec('xor') -> {400,400,500};
inop_prec('*') -> {500,500,600};
inop_prec('/') -> {500,500,600};
inop_prec('div') -> {500,500,600};
inop_prec('rem') -> {500,500,600};
inop_prec('band') -> {500,500,600};
inop_prec('and') -> {500,500,600};
inop_prec('#') -> {800,700,800};
inop_prec(':') -> {900,800,900}.

preop_prec('catch') -> {0,100};
preop_prec('+') -> {600,700};
preop_prec('-') -> {600,700};
preop_prec('bnot') -> {600,700};
preop_prec('not') -> {600,700};
preop_prec('#') -> {700,800}.

func_prec() -> {800,700}.

max_prec() -> 900.

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 2, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 260, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, dot, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 257, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(4, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, dot, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 250, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(function_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_function(__1),
 yeccpars2(yeccgoto(function, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(8, dot, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(9, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(rule_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_rule(__1),
 yeccpars2(yeccgoto(rule, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(11, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(13, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(rule_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [14 | __Ss], [__T | __Stack]);
yeccpars2(14, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(1,__1),
 yeccpars2(yeccgoto(clause_args, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [16 | __Ss], [__T | __Stack]);
yeccpars2(16, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(17, __Cat, [16 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [17 | __Ss], [__T | __Stack]);
yeccpars2(17, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(18, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [19 | __Ss], [__T | __Stack]);
yeccpars2(19, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(20, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(22, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 204, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [24 | __Ss], [__T | __Stack]);
yeccpars2(24, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(25, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(26, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(27, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(28, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(29, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(30, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [30 | __Ss], [__T | __Stack]);
yeccpars2(30, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 195, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, '!', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 191, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_100, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, 'orelse', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 189, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_150, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, 'andalso', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [38 | __Ss], [__T | __Stack]);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_160, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, '=/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '=:=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 180, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 183, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '>=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 178, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '=<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '/=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, '==', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [39 | __Ss], [__T | __Stack]);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_200, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, '--', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '++', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'xor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'or', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 173, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'bsr', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 170, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'bsl', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'bxor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'bor', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_300, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_400, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_500, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_600, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [45 | __Ss], [__T | __Stack]);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_800, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(guard, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(prefix_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(59, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [59 | __Ss], [__T | __Stack]);
yeccpars2(59, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(60, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [61 | __Ss], [__T | __Stack]);
yeccpars2(61, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_700, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [64 | __Ss], [__T | __Stack]);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(strings, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [66 | __Ss], [__T | __Stack]);
yeccpars2(66, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expr_max, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(71, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [71 | __Ss], [__T | __Stack]);
yeccpars2(71, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(72, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {tuple,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(75, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(76, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {try,line(__1),__2,[]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [77 | __Ss], [__T | __Stack]);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(cr_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [78 | __Ss], [__T | __Stack]);
yeccpars2(78, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(79, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(80, __Cat, [79 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [80 | __Ss], [__T | __Stack]);
yeccpars2(80, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(81, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(82, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),[__1],__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {try,line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(86, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cr_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {string,line(__1),element(3,__1) ++ element(3,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(strings, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 91, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(90, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(93, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),__2,__4,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(95, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(96, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(97, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'receive',line(__1),[],__3,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(98, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(99, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 100, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(100, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'query',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(query_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [102 | __Ss], [__T | __Stack]);
yeccpars2(102, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(103, '<-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(lc_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [104 | __Ss], [__T | __Stack]);
yeccpars2(104, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(lc_exprs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [105 | __Ss], [__T | __Stack]);
yeccpars2(105, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(106, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {lc,line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(list_comprehension, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {generate,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(lc_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr_600, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [112 | __Ss], [__T | __Stack]);
yeccpars2(112, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(113, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 116, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(if_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(115, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'if',line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(117, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(if_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(118, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(hd(hd(__1))),[],__1,__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(if_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [119 | __Ss], [__T | __Stack]);
yeccpars2(119, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(128, __Cat, [119 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, ';', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [121 | __Ss], [__T | __Stack]);
yeccpars2(121, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(fun_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 123, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_fun(line(__1),__2),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(125, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(126, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 127, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(127, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'fun',line(__1),{function,element(3,__2),element(3,__4)}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [128 | __Ss], [__T | __Stack]);
yeccpars2(128, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(129, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Args,Pos} = __1, {clause,Pos,'fun',Args,__2,__3}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(fun_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(130, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(131, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(guard, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [133 | __Ss], [__T | __Stack]);
yeccpars2(133, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(134, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(135, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {remote,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_800, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(136, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(137, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [137 | __Ss], [__T | __Stack]);
yeccpars2(137, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(138, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 153, [138 | __Ss], [__T | __Stack]);
yeccpars2(138, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(139, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__2),__1,element(3,__3),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(record_fields, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(143, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 148, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(145, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [146 | __Ss], [__T | __Stack]);
yeccpars2(146, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(147, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(148, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [149 | __Ss], [__T | __Stack]);
yeccpars2(149, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(150, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_fields, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(151, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(152, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__1),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_field, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_field,line(__2),__1,element(3,__3),__5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {call,line(__1),__1,element(1,__2)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(function_call, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(158, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(160, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(161, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mult_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_500, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(163, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(list_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(167, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [172 | __Ss], [__T | __Stack]);
yeccpars2(172, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(173, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(add_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_300, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, 'and', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, 'band', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 158, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, 'rem', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, 'div', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 155, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [176 | __Ss], [__T | __Stack]);
yeccpars2(176, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_400, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(181, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(comp_op, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(186, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_200, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_160, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(190, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_150, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [191 | __Ss], [__T | __Stack]);
yeccpars2(191, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(192, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {match,line(__2),__1,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_100, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(196, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(exprs, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'catch',line(__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(198, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(199, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(200, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(201, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {'case',line(__1),__2,__4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(case_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [202 | __Ss], [__T | __Stack]);
yeccpars2(202, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(203, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {block,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, '||', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 102, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 206, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(207, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,line(__1)},
 yeccpars2(yeccgoto(tail, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__1),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(210, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(212, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 206, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {cons,line(__2),__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),[]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 235, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bin_elements, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, '>>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 234, [216 | __Ss], [__T | __Stack]);
yeccpars2(216, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(217, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 221, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(222, __Cat, [217 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [219 | __Ss], [__T | __Stack]);
yeccpars2(219, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(220, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mkop(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(bit_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(221, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(222, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 223, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  default,
 yeccpars2(224, __Cat, [222 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(224, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin_element,line(__1),__1,__2,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_element, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 230, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  element(3,__1),
 yeccpars2(yeccgoto(bit_type, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(226, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(bit_type_list, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(228, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(229, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {element(3,__1),element(3,__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bit_type, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(opt_bit_size_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(233, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(bit_size_expr, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {bin,line(__1),__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(235, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [235 | __Ss], [__T | __Stack]);
yeccpars2(235, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(236, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(bin_elements, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(237, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 238, [237 | __Ss], [__T | __Stack]);
yeccpars2(237, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(238, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expr_max, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '.', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [240 | __Ss], [__T | __Stack]);
yeccpars2(240, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(241, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record,line(__1),element(3,__2),__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {record_index,line(__1),element(3,__2),__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(record_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(rule_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(rule_body, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(246, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {[],line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 248, [247 | __Ss], [__T | __Stack]);
yeccpars2(247, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(248, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,line(__1)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(argument_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(252, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(253, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [253 | __Ss], [__T | __Stack]);
yeccpars2(253, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(254, __Cat, [253 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(255, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {clause,line(__1),element(3,__1),__2,__3,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(function_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(257, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(form, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(259, __Cat, [258 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(259, ':-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 243, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 261, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(261, 'not', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'bnot', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '+', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, string, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, atom, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, float, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, integer, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, char, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'query', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, try, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'if', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '<<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'begin', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, var, __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(263, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(attr_val, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  build_attribute(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(add_op, 40) ->
 167;
yeccgoto(argument_list, 2) ->
 15;
yeccgoto(argument_list, 12) ->
 15;
yeccgoto(argument_list, 44) ->
 154;
yeccgoto(argument_list, 48) ->
 119;
yeccgoto(argument_list, 124) ->
 119;
yeccgoto(argument_list, 251) ->
 15;
yeccgoto(atomic, 14) ->
 26;
yeccgoto(atomic, 18) ->
 26;
yeccgoto(atomic, 20) ->
 26;
yeccgoto(atomic, 23) ->
 26;
yeccgoto(atomic, 24) ->
 26;
yeccgoto(atomic, 27) ->
 26;
yeccgoto(atomic, 30) ->
 26;
yeccgoto(atomic, 32) ->
 26;
yeccgoto(atomic, 52) ->
 26;
yeccgoto(atomic, 58) ->
 26;
yeccgoto(atomic, 61) ->
 26;
yeccgoto(atomic, 66) ->
 26;
yeccgoto(atomic, 70) ->
 26;
yeccgoto(atomic, 75) ->
 26;
yeccgoto(atomic, 81) ->
 26;
yeccgoto(atomic, 85) ->
 26;
yeccgoto(atomic, 88) ->
 26;
yeccgoto(atomic, 90) ->
 26;
yeccgoto(atomic, 98) ->
 26;
yeccgoto(atomic, 102) ->
 26;
yeccgoto(atomic, 107) ->
 26;
yeccgoto(atomic, 109) ->
 26;
yeccgoto(atomic, 116) ->
 26;
yeccgoto(atomic, 130) ->
 26;
yeccgoto(atomic, 134) ->
 26;
yeccgoto(atomic, 146) ->
 26;
yeccgoto(atomic, 151) ->
 26;
yeccgoto(atomic, 160) ->
 26;
yeccgoto(atomic, 167) ->
 26;
yeccgoto(atomic, 172) ->
 26;
yeccgoto(atomic, 185) ->
 26;
yeccgoto(atomic, 187) ->
 26;
yeccgoto(atomic, 189) ->
 26;
yeccgoto(atomic, 191) ->
 26;
yeccgoto(atomic, 192) ->
 26;
yeccgoto(atomic, 195) ->
 26;
yeccgoto(atomic, 199) ->
 26;
yeccgoto(atomic, 206) ->
 26;
yeccgoto(atomic, 209) ->
 26;
yeccgoto(atomic, 219) ->
 26;
yeccgoto(atomic, 221) ->
 26;
yeccgoto(atomic, 235) ->
 26;
yeccgoto(atomic, 243) ->
 26;
yeccgoto(atomic, 261) ->
 26;
yeccgoto(attr_val, 261) ->
 262;
yeccgoto(attribute, 0) ->
 3;
yeccgoto(bin_element, 23) ->
 215;
yeccgoto(bin_element, 235) ->
 215;
yeccgoto(bin_elements, 23) ->
 216;
yeccgoto(bin_elements, 235) ->
 236;
yeccgoto(binary, 14) ->
 28;
yeccgoto(binary, 18) ->
 28;
yeccgoto(binary, 20) ->
 28;
yeccgoto(binary, 23) ->
 28;
yeccgoto(binary, 24) ->
 28;
yeccgoto(binary, 27) ->
 28;
yeccgoto(binary, 30) ->
 28;
yeccgoto(binary, 32) ->
 28;
yeccgoto(binary, 52) ->
 28;
yeccgoto(binary, 58) ->
 28;
yeccgoto(binary, 61) ->
 28;
yeccgoto(binary, 66) ->
 28;
yeccgoto(binary, 70) ->
 28;
yeccgoto(binary, 75) ->
 28;
yeccgoto(binary, 81) ->
 28;
yeccgoto(binary, 85) ->
 28;
yeccgoto(binary, 88) ->
 28;
yeccgoto(binary, 90) ->
 28;
yeccgoto(binary, 98) ->
 28;
yeccgoto(binary, 102) ->
 28;
yeccgoto(binary, 107) ->
 28;
yeccgoto(binary, 109) ->
 28;
yeccgoto(binary, 116) ->
 28;
yeccgoto(binary, 130) ->
 28;
yeccgoto(binary, 134) ->
 28;
yeccgoto(binary, 146) ->
 28;
yeccgoto(binary, 151) ->
 28;
yeccgoto(binary, 160) ->
 28;
yeccgoto(binary, 167) ->
 28;
yeccgoto(binary, 172) ->
 28;
yeccgoto(binary, 185) ->
 28;
yeccgoto(binary, 187) ->
 28;
yeccgoto(binary, 189) ->
 28;
yeccgoto(binary, 191) ->
 28;
yeccgoto(binary, 192) ->
 28;
yeccgoto(binary, 195) ->
 28;
yeccgoto(binary, 199) ->
 28;
yeccgoto(binary, 206) ->
 28;
yeccgoto(binary, 209) ->
 28;
yeccgoto(binary, 219) ->
 28;
yeccgoto(binary, 221) ->
 28;
yeccgoto(binary, 235) ->
 28;
yeccgoto(binary, 243) ->
 28;
yeccgoto(binary, 261) ->
 28;
yeccgoto(bit_expr, 23) ->
 217;
yeccgoto(bit_expr, 235) ->
 217;
yeccgoto(bit_size_expr, 221) ->
 232;
yeccgoto(bit_type, 223) ->
 226;
yeccgoto(bit_type, 228) ->
 226;
yeccgoto(bit_type_list, 223) ->
 227;
yeccgoto(bit_type_list, 228) ->
 229;
yeccgoto(case_expr, 14) ->
 31;
yeccgoto(case_expr, 18) ->
 31;
yeccgoto(case_expr, 20) ->
 31;
yeccgoto(case_expr, 23) ->
 31;
yeccgoto(case_expr, 24) ->
 31;
yeccgoto(case_expr, 27) ->
 31;
yeccgoto(case_expr, 30) ->
 31;
yeccgoto(case_expr, 32) ->
 31;
yeccgoto(case_expr, 52) ->
 31;
yeccgoto(case_expr, 58) ->
 31;
yeccgoto(case_expr, 61) ->
 31;
yeccgoto(case_expr, 66) ->
 31;
yeccgoto(case_expr, 70) ->
 31;
yeccgoto(case_expr, 75) ->
 31;
yeccgoto(case_expr, 81) ->
 31;
yeccgoto(case_expr, 85) ->
 31;
yeccgoto(case_expr, 88) ->
 31;
yeccgoto(case_expr, 90) ->
 31;
yeccgoto(case_expr, 98) ->
 31;
yeccgoto(case_expr, 102) ->
 31;
yeccgoto(case_expr, 107) ->
 31;
yeccgoto(case_expr, 109) ->
 31;
yeccgoto(case_expr, 116) ->
 31;
yeccgoto(case_expr, 130) ->
 31;
yeccgoto(case_expr, 134) ->
 31;
yeccgoto(case_expr, 146) ->
 31;
yeccgoto(case_expr, 151) ->
 31;
yeccgoto(case_expr, 160) ->
 31;
yeccgoto(case_expr, 167) ->
 31;
yeccgoto(case_expr, 172) ->
 31;
yeccgoto(case_expr, 185) ->
 31;
yeccgoto(case_expr, 187) ->
 31;
yeccgoto(case_expr, 189) ->
 31;
yeccgoto(case_expr, 191) ->
 31;
yeccgoto(case_expr, 192) ->
 31;
yeccgoto(case_expr, 195) ->
 31;
yeccgoto(case_expr, 199) ->
 31;
yeccgoto(case_expr, 206) ->
 31;
yeccgoto(case_expr, 209) ->
 31;
yeccgoto(case_expr, 219) ->
 31;
yeccgoto(case_expr, 221) ->
 31;
yeccgoto(case_expr, 235) ->
 31;
yeccgoto(case_expr, 243) ->
 31;
yeccgoto(case_expr, 261) ->
 31;
yeccgoto(clause_args, 2) ->
 258;
yeccgoto(clause_args, 12) ->
 16;
yeccgoto(clause_args, 251) ->
 253;
yeccgoto(clause_body, 80) ->
 82;
yeccgoto(clause_body, 92) ->
 93;
yeccgoto(clause_body, 95) ->
 96;
yeccgoto(clause_body, 112) ->
 118;
yeccgoto(clause_body, 128) ->
 129;
yeccgoto(clause_body, 254) ->
 255;
yeccgoto(clause_body, 259) ->
 255;
yeccgoto(clause_guard, 16) ->
 17;
yeccgoto(clause_guard, 79) ->
 80;
yeccgoto(clause_guard, 119) ->
 128;
yeccgoto(clause_guard, 253) ->
 254;
yeccgoto(clause_guard, 258) ->
 259;
yeccgoto(comp_op, 39) ->
 185;
yeccgoto(cr_clause, 61) ->
 77;
yeccgoto(cr_clause, 75) ->
 77;
yeccgoto(cr_clause, 85) ->
 77;
yeccgoto(cr_clause, 199) ->
 77;
yeccgoto(cr_clauses, 61) ->
 89;
yeccgoto(cr_clauses, 75) ->
 78;
yeccgoto(cr_clauses, 85) ->
 86;
yeccgoto(cr_clauses, 199) ->
 200;
yeccgoto(expr, 14) ->
 34;
yeccgoto(expr, 18) ->
 34;
yeccgoto(expr, 20) ->
 237;
yeccgoto(expr, 24) ->
 205;
yeccgoto(expr, 27) ->
 34;
yeccgoto(expr, 30) ->
 198;
yeccgoto(expr, 32) ->
 197;
yeccgoto(expr, 52) ->
 34;
yeccgoto(expr, 61) ->
 79;
yeccgoto(expr, 66) ->
 34;
yeccgoto(expr, 70) ->
 34;
yeccgoto(expr, 75) ->
 79;
yeccgoto(expr, 81) ->
 34;
yeccgoto(expr, 85) ->
 79;
yeccgoto(expr, 88) ->
 95;
yeccgoto(expr, 90) ->
 92;
yeccgoto(expr, 98) ->
 101;
yeccgoto(expr, 102) ->
 103;
yeccgoto(expr, 107) ->
 103;
yeccgoto(expr, 109) ->
 110;
yeccgoto(expr, 116) ->
 34;
yeccgoto(expr, 130) ->
 34;
yeccgoto(expr, 146) ->
 147;
yeccgoto(expr, 151) ->
 152;
yeccgoto(expr, 195) ->
 34;
yeccgoto(expr, 199) ->
 79;
yeccgoto(expr, 206) ->
 212;
yeccgoto(expr, 209) ->
 210;
yeccgoto(expr, 243) ->
 103;
yeccgoto(expr, 261) ->
 34;
yeccgoto(expr_100, 14) ->
 35;
yeccgoto(expr_100, 18) ->
 35;
yeccgoto(expr_100, 20) ->
 35;
yeccgoto(expr_100, 24) ->
 35;
yeccgoto(expr_100, 27) ->
 35;
yeccgoto(expr_100, 30) ->
 35;
yeccgoto(expr_100, 32) ->
 35;
yeccgoto(expr_100, 52) ->
 35;
yeccgoto(expr_100, 61) ->
 35;
yeccgoto(expr_100, 66) ->
 35;
yeccgoto(expr_100, 70) ->
 35;
yeccgoto(expr_100, 75) ->
 35;
yeccgoto(expr_100, 81) ->
 35;
yeccgoto(expr_100, 85) ->
 35;
yeccgoto(expr_100, 88) ->
 35;
yeccgoto(expr_100, 90) ->
 35;
yeccgoto(expr_100, 98) ->
 35;
yeccgoto(expr_100, 102) ->
 35;
yeccgoto(expr_100, 107) ->
 35;
yeccgoto(expr_100, 109) ->
 35;
yeccgoto(expr_100, 116) ->
 35;
yeccgoto(expr_100, 130) ->
 35;
yeccgoto(expr_100, 146) ->
 35;
yeccgoto(expr_100, 151) ->
 35;
yeccgoto(expr_100, 191) ->
 194;
yeccgoto(expr_100, 192) ->
 193;
yeccgoto(expr_100, 195) ->
 35;
yeccgoto(expr_100, 199) ->
 35;
yeccgoto(expr_100, 206) ->
 35;
yeccgoto(expr_100, 209) ->
 35;
yeccgoto(expr_100, 243) ->
 35;
yeccgoto(expr_100, 261) ->
 35;
yeccgoto(expr_150, 14) ->
 36;
yeccgoto(expr_150, 18) ->
 36;
yeccgoto(expr_150, 20) ->
 36;
yeccgoto(expr_150, 24) ->
 36;
yeccgoto(expr_150, 27) ->
 36;
yeccgoto(expr_150, 30) ->
 36;
yeccgoto(expr_150, 32) ->
 36;
yeccgoto(expr_150, 52) ->
 36;
yeccgoto(expr_150, 61) ->
 36;
yeccgoto(expr_150, 66) ->
 36;
yeccgoto(expr_150, 70) ->
 36;
yeccgoto(expr_150, 75) ->
 36;
yeccgoto(expr_150, 81) ->
 36;
yeccgoto(expr_150, 85) ->
 36;
yeccgoto(expr_150, 88) ->
 36;
yeccgoto(expr_150, 90) ->
 36;
yeccgoto(expr_150, 98) ->
 36;
yeccgoto(expr_150, 102) ->
 36;
yeccgoto(expr_150, 107) ->
 36;
yeccgoto(expr_150, 109) ->
 36;
yeccgoto(expr_150, 116) ->
 36;
yeccgoto(expr_150, 130) ->
 36;
yeccgoto(expr_150, 146) ->
 36;
yeccgoto(expr_150, 151) ->
 36;
yeccgoto(expr_150, 189) ->
 190;
yeccgoto(expr_150, 191) ->
 36;
yeccgoto(expr_150, 192) ->
 36;
yeccgoto(expr_150, 195) ->
 36;
yeccgoto(expr_150, 199) ->
 36;
yeccgoto(expr_150, 206) ->
 36;
yeccgoto(expr_150, 209) ->
 36;
yeccgoto(expr_150, 243) ->
 36;
yeccgoto(expr_150, 261) ->
 36;
yeccgoto(expr_160, 14) ->
 37;
yeccgoto(expr_160, 18) ->
 37;
yeccgoto(expr_160, 20) ->
 37;
yeccgoto(expr_160, 24) ->
 37;
yeccgoto(expr_160, 27) ->
 37;
yeccgoto(expr_160, 30) ->
 37;
yeccgoto(expr_160, 32) ->
 37;
yeccgoto(expr_160, 52) ->
 37;
yeccgoto(expr_160, 61) ->
 37;
yeccgoto(expr_160, 66) ->
 37;
yeccgoto(expr_160, 70) ->
 37;
yeccgoto(expr_160, 75) ->
 37;
yeccgoto(expr_160, 81) ->
 37;
yeccgoto(expr_160, 85) ->
 37;
yeccgoto(expr_160, 88) ->
 37;
yeccgoto(expr_160, 90) ->
 37;
yeccgoto(expr_160, 98) ->
 37;
yeccgoto(expr_160, 102) ->
 37;
yeccgoto(expr_160, 107) ->
 37;
yeccgoto(expr_160, 109) ->
 37;
yeccgoto(expr_160, 116) ->
 37;
yeccgoto(expr_160, 130) ->
 37;
yeccgoto(expr_160, 146) ->
 37;
yeccgoto(expr_160, 151) ->
 37;
yeccgoto(expr_160, 187) ->
 188;
yeccgoto(expr_160, 189) ->
 37;
yeccgoto(expr_160, 191) ->
 37;
yeccgoto(expr_160, 192) ->
 37;
yeccgoto(expr_160, 195) ->
 37;
yeccgoto(expr_160, 199) ->
 37;
yeccgoto(expr_160, 206) ->
 37;
yeccgoto(expr_160, 209) ->
 37;
yeccgoto(expr_160, 243) ->
 37;
yeccgoto(expr_160, 261) ->
 37;
yeccgoto(expr_200, 14) ->
 38;
yeccgoto(expr_200, 18) ->
 38;
yeccgoto(expr_200, 20) ->
 38;
yeccgoto(expr_200, 24) ->
 38;
yeccgoto(expr_200, 27) ->
 38;
yeccgoto(expr_200, 30) ->
 38;
yeccgoto(expr_200, 32) ->
 38;
yeccgoto(expr_200, 52) ->
 38;
yeccgoto(expr_200, 61) ->
 38;
yeccgoto(expr_200, 66) ->
 38;
yeccgoto(expr_200, 70) ->
 38;
yeccgoto(expr_200, 75) ->
 38;
yeccgoto(expr_200, 81) ->
 38;
yeccgoto(expr_200, 85) ->
 38;
yeccgoto(expr_200, 88) ->
 38;
yeccgoto(expr_200, 90) ->
 38;
yeccgoto(expr_200, 98) ->
 38;
yeccgoto(expr_200, 102) ->
 38;
yeccgoto(expr_200, 107) ->
 38;
yeccgoto(expr_200, 109) ->
 38;
yeccgoto(expr_200, 116) ->
 38;
yeccgoto(expr_200, 130) ->
 38;
yeccgoto(expr_200, 146) ->
 38;
yeccgoto(expr_200, 151) ->
 38;
yeccgoto(expr_200, 187) ->
 38;
yeccgoto(expr_200, 189) ->
 38;
yeccgoto(expr_200, 191) ->
 38;
yeccgoto(expr_200, 192) ->
 38;
yeccgoto(expr_200, 195) ->
 38;
yeccgoto(expr_200, 199) ->
 38;
yeccgoto(expr_200, 206) ->
 38;
yeccgoto(expr_200, 209) ->
 38;
yeccgoto(expr_200, 243) ->
 38;
yeccgoto(expr_200, 261) ->
 38;
yeccgoto(expr_300, 14) ->
 39;
yeccgoto(expr_300, 18) ->
 39;
yeccgoto(expr_300, 20) ->
 39;
yeccgoto(expr_300, 24) ->
 39;
yeccgoto(expr_300, 27) ->
 39;
yeccgoto(expr_300, 30) ->
 39;
yeccgoto(expr_300, 32) ->
 39;
yeccgoto(expr_300, 52) ->
 39;
yeccgoto(expr_300, 61) ->
 39;
yeccgoto(expr_300, 66) ->
 39;
yeccgoto(expr_300, 70) ->
 39;
yeccgoto(expr_300, 75) ->
 39;
yeccgoto(expr_300, 81) ->
 39;
yeccgoto(expr_300, 85) ->
 39;
yeccgoto(expr_300, 88) ->
 39;
yeccgoto(expr_300, 90) ->
 39;
yeccgoto(expr_300, 98) ->
 39;
yeccgoto(expr_300, 102) ->
 39;
yeccgoto(expr_300, 107) ->
 39;
yeccgoto(expr_300, 109) ->
 39;
yeccgoto(expr_300, 116) ->
 39;
yeccgoto(expr_300, 130) ->
 39;
yeccgoto(expr_300, 146) ->
 39;
yeccgoto(expr_300, 151) ->
 39;
yeccgoto(expr_300, 172) ->
 175;
yeccgoto(expr_300, 185) ->
 186;
yeccgoto(expr_300, 187) ->
 39;
yeccgoto(expr_300, 189) ->
 39;
yeccgoto(expr_300, 191) ->
 39;
yeccgoto(expr_300, 192) ->
 39;
yeccgoto(expr_300, 195) ->
 39;
yeccgoto(expr_300, 199) ->
 39;
yeccgoto(expr_300, 206) ->
 39;
yeccgoto(expr_300, 209) ->
 39;
yeccgoto(expr_300, 243) ->
 39;
yeccgoto(expr_300, 261) ->
 39;
yeccgoto(expr_400, 14) ->
 40;
yeccgoto(expr_400, 18) ->
 40;
yeccgoto(expr_400, 20) ->
 40;
yeccgoto(expr_400, 24) ->
 40;
yeccgoto(expr_400, 27) ->
 40;
yeccgoto(expr_400, 30) ->
 40;
yeccgoto(expr_400, 32) ->
 40;
yeccgoto(expr_400, 52) ->
 40;
yeccgoto(expr_400, 61) ->
 40;
yeccgoto(expr_400, 66) ->
 40;
yeccgoto(expr_400, 70) ->
 40;
yeccgoto(expr_400, 75) ->
 40;
yeccgoto(expr_400, 81) ->
 40;
yeccgoto(expr_400, 85) ->
 40;
yeccgoto(expr_400, 88) ->
 40;
yeccgoto(expr_400, 90) ->
 40;
yeccgoto(expr_400, 98) ->
 40;
yeccgoto(expr_400, 102) ->
 40;
yeccgoto(expr_400, 107) ->
 40;
yeccgoto(expr_400, 109) ->
 40;
yeccgoto(expr_400, 116) ->
 40;
yeccgoto(expr_400, 130) ->
 40;
yeccgoto(expr_400, 146) ->
 40;
yeccgoto(expr_400, 151) ->
 40;
yeccgoto(expr_400, 172) ->
 40;
yeccgoto(expr_400, 185) ->
 40;
yeccgoto(expr_400, 187) ->
 40;
yeccgoto(expr_400, 189) ->
 40;
yeccgoto(expr_400, 191) ->
 40;
yeccgoto(expr_400, 192) ->
 40;
yeccgoto(expr_400, 195) ->
 40;
yeccgoto(expr_400, 199) ->
 40;
yeccgoto(expr_400, 206) ->
 40;
yeccgoto(expr_400, 209) ->
 40;
yeccgoto(expr_400, 243) ->
 40;
yeccgoto(expr_400, 261) ->
 40;
yeccgoto(expr_500, 14) ->
 41;
yeccgoto(expr_500, 18) ->
 41;
yeccgoto(expr_500, 20) ->
 41;
yeccgoto(expr_500, 24) ->
 41;
yeccgoto(expr_500, 27) ->
 41;
yeccgoto(expr_500, 30) ->
 41;
yeccgoto(expr_500, 32) ->
 41;
yeccgoto(expr_500, 52) ->
 41;
yeccgoto(expr_500, 61) ->
 41;
yeccgoto(expr_500, 66) ->
 41;
yeccgoto(expr_500, 70) ->
 41;
yeccgoto(expr_500, 75) ->
 41;
yeccgoto(expr_500, 81) ->
 41;
yeccgoto(expr_500, 85) ->
 41;
yeccgoto(expr_500, 88) ->
 41;
yeccgoto(expr_500, 90) ->
 41;
yeccgoto(expr_500, 98) ->
 41;
yeccgoto(expr_500, 102) ->
 41;
yeccgoto(expr_500, 107) ->
 41;
yeccgoto(expr_500, 109) ->
 41;
yeccgoto(expr_500, 116) ->
 41;
yeccgoto(expr_500, 130) ->
 41;
yeccgoto(expr_500, 146) ->
 41;
yeccgoto(expr_500, 151) ->
 41;
yeccgoto(expr_500, 167) ->
 176;
yeccgoto(expr_500, 172) ->
 41;
yeccgoto(expr_500, 185) ->
 41;
yeccgoto(expr_500, 187) ->
 41;
yeccgoto(expr_500, 189) ->
 41;
yeccgoto(expr_500, 191) ->
 41;
yeccgoto(expr_500, 192) ->
 41;
yeccgoto(expr_500, 195) ->
 41;
yeccgoto(expr_500, 199) ->
 41;
yeccgoto(expr_500, 206) ->
 41;
yeccgoto(expr_500, 209) ->
 41;
yeccgoto(expr_500, 243) ->
 41;
yeccgoto(expr_500, 261) ->
 41;
yeccgoto(expr_600, 14) ->
 42;
yeccgoto(expr_600, 18) ->
 42;
yeccgoto(expr_600, 20) ->
 42;
yeccgoto(expr_600, 24) ->
 42;
yeccgoto(expr_600, 27) ->
 42;
yeccgoto(expr_600, 30) ->
 42;
yeccgoto(expr_600, 32) ->
 42;
yeccgoto(expr_600, 52) ->
 42;
yeccgoto(expr_600, 61) ->
 42;
yeccgoto(expr_600, 66) ->
 42;
yeccgoto(expr_600, 70) ->
 42;
yeccgoto(expr_600, 75) ->
 42;
yeccgoto(expr_600, 81) ->
 42;
yeccgoto(expr_600, 85) ->
 42;
yeccgoto(expr_600, 88) ->
 42;
yeccgoto(expr_600, 90) ->
 42;
yeccgoto(expr_600, 98) ->
 42;
yeccgoto(expr_600, 102) ->
 42;
yeccgoto(expr_600, 107) ->
 42;
yeccgoto(expr_600, 109) ->
 42;
yeccgoto(expr_600, 116) ->
 42;
yeccgoto(expr_600, 130) ->
 42;
yeccgoto(expr_600, 146) ->
 42;
yeccgoto(expr_600, 151) ->
 42;
yeccgoto(expr_600, 160) ->
 162;
yeccgoto(expr_600, 167) ->
 42;
yeccgoto(expr_600, 172) ->
 42;
yeccgoto(expr_600, 185) ->
 42;
yeccgoto(expr_600, 187) ->
 42;
yeccgoto(expr_600, 189) ->
 42;
yeccgoto(expr_600, 191) ->
 42;
yeccgoto(expr_600, 192) ->
 42;
yeccgoto(expr_600, 195) ->
 42;
yeccgoto(expr_600, 199) ->
 42;
yeccgoto(expr_600, 206) ->
 42;
yeccgoto(expr_600, 209) ->
 42;
yeccgoto(expr_600, 243) ->
 42;
yeccgoto(expr_600, 261) ->
 42;
yeccgoto(expr_700, 14) ->
 43;
yeccgoto(expr_700, 18) ->
 43;
yeccgoto(expr_700, 20) ->
 43;
yeccgoto(expr_700, 24) ->
 43;
yeccgoto(expr_700, 27) ->
 43;
yeccgoto(expr_700, 30) ->
 43;
yeccgoto(expr_700, 32) ->
 43;
yeccgoto(expr_700, 52) ->
 43;
yeccgoto(expr_700, 58) ->
 111;
yeccgoto(expr_700, 61) ->
 43;
yeccgoto(expr_700, 66) ->
 43;
yeccgoto(expr_700, 70) ->
 43;
yeccgoto(expr_700, 75) ->
 43;
yeccgoto(expr_700, 81) ->
 43;
yeccgoto(expr_700, 85) ->
 43;
yeccgoto(expr_700, 88) ->
 43;
yeccgoto(expr_700, 90) ->
 43;
yeccgoto(expr_700, 98) ->
 43;
yeccgoto(expr_700, 102) ->
 43;
yeccgoto(expr_700, 107) ->
 43;
yeccgoto(expr_700, 109) ->
 43;
yeccgoto(expr_700, 116) ->
 43;
yeccgoto(expr_700, 130) ->
 43;
yeccgoto(expr_700, 146) ->
 43;
yeccgoto(expr_700, 151) ->
 43;
yeccgoto(expr_700, 160) ->
 43;
yeccgoto(expr_700, 167) ->
 43;
yeccgoto(expr_700, 172) ->
 43;
yeccgoto(expr_700, 185) ->
 43;
yeccgoto(expr_700, 187) ->
 43;
yeccgoto(expr_700, 189) ->
 43;
yeccgoto(expr_700, 191) ->
 43;
yeccgoto(expr_700, 192) ->
 43;
yeccgoto(expr_700, 195) ->
 43;
yeccgoto(expr_700, 199) ->
 43;
yeccgoto(expr_700, 206) ->
 43;
yeccgoto(expr_700, 209) ->
 43;
yeccgoto(expr_700, 243) ->
 43;
yeccgoto(expr_700, 261) ->
 43;
yeccgoto(expr_800, 14) ->
 44;
yeccgoto(expr_800, 18) ->
 44;
yeccgoto(expr_800, 20) ->
 44;
yeccgoto(expr_800, 24) ->
 44;
yeccgoto(expr_800, 27) ->
 44;
yeccgoto(expr_800, 30) ->
 44;
yeccgoto(expr_800, 32) ->
 44;
yeccgoto(expr_800, 52) ->
 44;
yeccgoto(expr_800, 58) ->
 44;
yeccgoto(expr_800, 61) ->
 44;
yeccgoto(expr_800, 66) ->
 44;
yeccgoto(expr_800, 70) ->
 44;
yeccgoto(expr_800, 75) ->
 44;
yeccgoto(expr_800, 81) ->
 44;
yeccgoto(expr_800, 85) ->
 44;
yeccgoto(expr_800, 88) ->
 44;
yeccgoto(expr_800, 90) ->
 44;
yeccgoto(expr_800, 98) ->
 44;
yeccgoto(expr_800, 102) ->
 44;
yeccgoto(expr_800, 107) ->
 44;
yeccgoto(expr_800, 109) ->
 44;
yeccgoto(expr_800, 116) ->
 44;
yeccgoto(expr_800, 130) ->
 44;
yeccgoto(expr_800, 146) ->
 44;
yeccgoto(expr_800, 151) ->
 44;
yeccgoto(expr_800, 160) ->
 44;
yeccgoto(expr_800, 167) ->
 44;
yeccgoto(expr_800, 172) ->
 44;
yeccgoto(expr_800, 185) ->
 44;
yeccgoto(expr_800, 187) ->
 44;
yeccgoto(expr_800, 189) ->
 44;
yeccgoto(expr_800, 191) ->
 44;
yeccgoto(expr_800, 192) ->
 44;
yeccgoto(expr_800, 195) ->
 44;
yeccgoto(expr_800, 199) ->
 44;
yeccgoto(expr_800, 206) ->
 44;
yeccgoto(expr_800, 209) ->
 44;
yeccgoto(expr_800, 243) ->
 44;
yeccgoto(expr_800, 261) ->
 44;
yeccgoto(expr_max, 14) ->
 45;
yeccgoto(expr_max, 18) ->
 45;
yeccgoto(expr_max, 20) ->
 45;
yeccgoto(expr_max, 23) ->
 218;
yeccgoto(expr_max, 24) ->
 45;
yeccgoto(expr_max, 27) ->
 45;
yeccgoto(expr_max, 30) ->
 45;
yeccgoto(expr_max, 32) ->
 45;
yeccgoto(expr_max, 52) ->
 45;
yeccgoto(expr_max, 58) ->
 45;
yeccgoto(expr_max, 61) ->
 45;
yeccgoto(expr_max, 66) ->
 45;
yeccgoto(expr_max, 70) ->
 45;
yeccgoto(expr_max, 75) ->
 45;
yeccgoto(expr_max, 81) ->
 45;
yeccgoto(expr_max, 85) ->
 45;
yeccgoto(expr_max, 88) ->
 45;
yeccgoto(expr_max, 90) ->
 45;
yeccgoto(expr_max, 98) ->
 45;
yeccgoto(expr_max, 102) ->
 45;
yeccgoto(expr_max, 107) ->
 45;
yeccgoto(expr_max, 109) ->
 45;
yeccgoto(expr_max, 116) ->
 45;
yeccgoto(expr_max, 130) ->
 45;
yeccgoto(expr_max, 134) ->
 135;
yeccgoto(expr_max, 146) ->
 45;
yeccgoto(expr_max, 151) ->
 45;
yeccgoto(expr_max, 160) ->
 45;
yeccgoto(expr_max, 167) ->
 45;
yeccgoto(expr_max, 172) ->
 45;
yeccgoto(expr_max, 185) ->
 45;
yeccgoto(expr_max, 187) ->
 45;
yeccgoto(expr_max, 189) ->
 45;
yeccgoto(expr_max, 191) ->
 45;
yeccgoto(expr_max, 192) ->
 45;
yeccgoto(expr_max, 195) ->
 45;
yeccgoto(expr_max, 199) ->
 45;
yeccgoto(expr_max, 206) ->
 45;
yeccgoto(expr_max, 209) ->
 45;
yeccgoto(expr_max, 219) ->
 220;
yeccgoto(expr_max, 221) ->
 233;
yeccgoto(expr_max, 235) ->
 218;
yeccgoto(expr_max, 243) ->
 45;
yeccgoto(expr_max, 261) ->
 45;
yeccgoto(exprs, 14) ->
 247;
yeccgoto(exprs, 18) ->
 46;
yeccgoto(exprs, 27) ->
 202;
yeccgoto(exprs, 52) ->
 46;
yeccgoto(exprs, 66) ->
 74;
yeccgoto(exprs, 70) ->
 71;
yeccgoto(exprs, 81) ->
 83;
yeccgoto(exprs, 116) ->
 46;
yeccgoto(exprs, 130) ->
 46;
yeccgoto(exprs, 195) ->
 196;
yeccgoto(exprs, 261) ->
 263;
yeccgoto(form, 0) ->
 4;
yeccgoto(fun_clause, 48) ->
 121;
yeccgoto(fun_clause, 124) ->
 121;
yeccgoto(fun_clauses, 48) ->
 122;
yeccgoto(fun_clauses, 124) ->
 125;
yeccgoto(fun_expr, 14) ->
 49;
yeccgoto(fun_expr, 18) ->
 49;
yeccgoto(fun_expr, 20) ->
 49;
yeccgoto(fun_expr, 23) ->
 49;
yeccgoto(fun_expr, 24) ->
 49;
yeccgoto(fun_expr, 27) ->
 49;
yeccgoto(fun_expr, 30) ->
 49;
yeccgoto(fun_expr, 32) ->
 49;
yeccgoto(fun_expr, 52) ->
 49;
yeccgoto(fun_expr, 58) ->
 49;
yeccgoto(fun_expr, 61) ->
 49;
yeccgoto(fun_expr, 66) ->
 49;
yeccgoto(fun_expr, 70) ->
 49;
yeccgoto(fun_expr, 75) ->
 49;
yeccgoto(fun_expr, 81) ->
 49;
yeccgoto(fun_expr, 85) ->
 49;
yeccgoto(fun_expr, 88) ->
 49;
yeccgoto(fun_expr, 90) ->
 49;
yeccgoto(fun_expr, 98) ->
 49;
yeccgoto(fun_expr, 102) ->
 49;
yeccgoto(fun_expr, 107) ->
 49;
yeccgoto(fun_expr, 109) ->
 49;
yeccgoto(fun_expr, 116) ->
 49;
yeccgoto(fun_expr, 130) ->
 49;
yeccgoto(fun_expr, 134) ->
 49;
yeccgoto(fun_expr, 146) ->
 49;
yeccgoto(fun_expr, 151) ->
 49;
yeccgoto(fun_expr, 160) ->
 49;
yeccgoto(fun_expr, 167) ->
 49;
yeccgoto(fun_expr, 172) ->
 49;
yeccgoto(fun_expr, 185) ->
 49;
yeccgoto(fun_expr, 187) ->
 49;
yeccgoto(fun_expr, 189) ->
 49;
yeccgoto(fun_expr, 191) ->
 49;
yeccgoto(fun_expr, 192) ->
 49;
yeccgoto(fun_expr, 195) ->
 49;
yeccgoto(fun_expr, 199) ->
 49;
yeccgoto(fun_expr, 206) ->
 49;
yeccgoto(fun_expr, 209) ->
 49;
yeccgoto(fun_expr, 219) ->
 49;
yeccgoto(fun_expr, 221) ->
 49;
yeccgoto(fun_expr, 235) ->
 49;
yeccgoto(fun_expr, 243) ->
 49;
yeccgoto(fun_expr, 261) ->
 49;
yeccgoto(function, 0) ->
 5;
yeccgoto(function_call, 14) ->
 50;
yeccgoto(function_call, 18) ->
 50;
yeccgoto(function_call, 20) ->
 50;
yeccgoto(function_call, 24) ->
 50;
yeccgoto(function_call, 27) ->
 50;
yeccgoto(function_call, 30) ->
 50;
yeccgoto(function_call, 32) ->
 50;
yeccgoto(function_call, 52) ->
 50;
yeccgoto(function_call, 58) ->
 50;
yeccgoto(function_call, 61) ->
 50;
yeccgoto(function_call, 66) ->
 50;
yeccgoto(function_call, 70) ->
 50;
yeccgoto(function_call, 75) ->
 50;
yeccgoto(function_call, 81) ->
 50;
yeccgoto(function_call, 85) ->
 50;
yeccgoto(function_call, 88) ->
 50;
yeccgoto(function_call, 90) ->
 50;
yeccgoto(function_call, 98) ->
 50;
yeccgoto(function_call, 102) ->
 50;
yeccgoto(function_call, 107) ->
 50;
yeccgoto(function_call, 109) ->
 50;
yeccgoto(function_call, 116) ->
 50;
yeccgoto(function_call, 130) ->
 50;
yeccgoto(function_call, 146) ->
 50;
yeccgoto(function_call, 151) ->
 50;
yeccgoto(function_call, 160) ->
 50;
yeccgoto(function_call, 167) ->
 50;
yeccgoto(function_call, 172) ->
 50;
yeccgoto(function_call, 185) ->
 50;
yeccgoto(function_call, 187) ->
 50;
yeccgoto(function_call, 189) ->
 50;
yeccgoto(function_call, 191) ->
 50;
yeccgoto(function_call, 192) ->
 50;
yeccgoto(function_call, 195) ->
 50;
yeccgoto(function_call, 199) ->
 50;
yeccgoto(function_call, 206) ->
 50;
yeccgoto(function_call, 209) ->
 50;
yeccgoto(function_call, 243) ->
 50;
yeccgoto(function_call, 261) ->
 50;
yeccgoto(function_clause, 0) ->
 6;
yeccgoto(function_clause, 250) ->
 6;
yeccgoto(function_clauses, 0) ->
 7;
yeccgoto(function_clauses, 250) ->
 252;
yeccgoto(guard, 18) ->
 51;
yeccgoto(guard, 52) ->
 112;
yeccgoto(guard, 116) ->
 112;
yeccgoto(guard, 130) ->
 131;
yeccgoto(if_clause, 52) ->
 113;
yeccgoto(if_clause, 116) ->
 113;
yeccgoto(if_clauses, 52) ->
 114;
yeccgoto(if_clauses, 116) ->
 117;
yeccgoto(if_expr, 14) ->
 53;
yeccgoto(if_expr, 18) ->
 53;
yeccgoto(if_expr, 20) ->
 53;
yeccgoto(if_expr, 23) ->
 53;
yeccgoto(if_expr, 24) ->
 53;
yeccgoto(if_expr, 27) ->
 53;
yeccgoto(if_expr, 30) ->
 53;
yeccgoto(if_expr, 32) ->
 53;
yeccgoto(if_expr, 52) ->
 53;
yeccgoto(if_expr, 58) ->
 53;
yeccgoto(if_expr, 61) ->
 53;
yeccgoto(if_expr, 66) ->
 53;
yeccgoto(if_expr, 70) ->
 53;
yeccgoto(if_expr, 75) ->
 53;
yeccgoto(if_expr, 81) ->
 53;
yeccgoto(if_expr, 85) ->
 53;
yeccgoto(if_expr, 88) ->
 53;
yeccgoto(if_expr, 90) ->
 53;
yeccgoto(if_expr, 98) ->
 53;
yeccgoto(if_expr, 102) ->
 53;
yeccgoto(if_expr, 107) ->
 53;
yeccgoto(if_expr, 109) ->
 53;
yeccgoto(if_expr, 116) ->
 53;
yeccgoto(if_expr, 130) ->
 53;
yeccgoto(if_expr, 134) ->
 53;
yeccgoto(if_expr, 146) ->
 53;
yeccgoto(if_expr, 151) ->
 53;
yeccgoto(if_expr, 160) ->
 53;
yeccgoto(if_expr, 167) ->
 53;
yeccgoto(if_expr, 172) ->
 53;
yeccgoto(if_expr, 185) ->
 53;
yeccgoto(if_expr, 187) ->
 53;
yeccgoto(if_expr, 189) ->
 53;
yeccgoto(if_expr, 191) ->
 53;
yeccgoto(if_expr, 192) ->
 53;
yeccgoto(if_expr, 195) ->
 53;
yeccgoto(if_expr, 199) ->
 53;
yeccgoto(if_expr, 206) ->
 53;
yeccgoto(if_expr, 209) ->
 53;
yeccgoto(if_expr, 219) ->
 53;
yeccgoto(if_expr, 221) ->
 53;
yeccgoto(if_expr, 235) ->
 53;
yeccgoto(if_expr, 243) ->
 53;
yeccgoto(if_expr, 261) ->
 53;
yeccgoto(lc_expr, 102) ->
 104;
yeccgoto(lc_expr, 107) ->
 104;
yeccgoto(lc_expr, 243) ->
 104;
yeccgoto(lc_exprs, 102) ->
 105;
yeccgoto(lc_exprs, 107) ->
 108;
yeccgoto(lc_exprs, 243) ->
 245;
yeccgoto(list, 14) ->
 55;
yeccgoto(list, 18) ->
 55;
yeccgoto(list, 20) ->
 55;
yeccgoto(list, 23) ->
 55;
yeccgoto(list, 24) ->
 55;
yeccgoto(list, 27) ->
 55;
yeccgoto(list, 30) ->
 55;
yeccgoto(list, 32) ->
 55;
yeccgoto(list, 52) ->
 55;
yeccgoto(list, 58) ->
 55;
yeccgoto(list, 61) ->
 55;
yeccgoto(list, 66) ->
 55;
yeccgoto(list, 70) ->
 55;
yeccgoto(list, 75) ->
 55;
yeccgoto(list, 81) ->
 55;
yeccgoto(list, 85) ->
 55;
yeccgoto(list, 88) ->
 55;
yeccgoto(list, 90) ->
 55;
yeccgoto(list, 98) ->
 55;
yeccgoto(list, 102) ->
 55;
yeccgoto(list, 107) ->
 55;
yeccgoto(list, 109) ->
 55;
yeccgoto(list, 116) ->
 55;
yeccgoto(list, 130) ->
 55;
yeccgoto(list, 134) ->
 55;
yeccgoto(list, 146) ->
 55;
yeccgoto(list, 151) ->
 55;
yeccgoto(list, 160) ->
 55;
yeccgoto(list, 167) ->
 55;
yeccgoto(list, 172) ->
 55;
yeccgoto(list, 185) ->
 55;
yeccgoto(list, 187) ->
 55;
yeccgoto(list, 189) ->
 55;
yeccgoto(list, 191) ->
 55;
yeccgoto(list, 192) ->
 55;
yeccgoto(list, 195) ->
 55;
yeccgoto(list, 199) ->
 55;
yeccgoto(list, 206) ->
 55;
yeccgoto(list, 209) ->
 55;
yeccgoto(list, 219) ->
 55;
yeccgoto(list, 221) ->
 55;
yeccgoto(list, 235) ->
 55;
yeccgoto(list, 243) ->
 55;
yeccgoto(list, 261) ->
 55;
yeccgoto(list_comprehension, 14) ->
 56;
yeccgoto(list_comprehension, 18) ->
 56;
yeccgoto(list_comprehension, 20) ->
 56;
yeccgoto(list_comprehension, 23) ->
 56;
yeccgoto(list_comprehension, 24) ->
 56;
yeccgoto(list_comprehension, 27) ->
 56;
yeccgoto(list_comprehension, 30) ->
 56;
yeccgoto(list_comprehension, 32) ->
 56;
yeccgoto(list_comprehension, 52) ->
 56;
yeccgoto(list_comprehension, 58) ->
 56;
yeccgoto(list_comprehension, 59) ->
 99;
yeccgoto(list_comprehension, 61) ->
 56;
yeccgoto(list_comprehension, 66) ->
 56;
yeccgoto(list_comprehension, 70) ->
 56;
yeccgoto(list_comprehension, 75) ->
 56;
yeccgoto(list_comprehension, 81) ->
 56;
yeccgoto(list_comprehension, 85) ->
 56;
yeccgoto(list_comprehension, 88) ->
 56;
yeccgoto(list_comprehension, 90) ->
 56;
yeccgoto(list_comprehension, 98) ->
 56;
yeccgoto(list_comprehension, 102) ->
 56;
yeccgoto(list_comprehension, 107) ->
 56;
yeccgoto(list_comprehension, 109) ->
 56;
yeccgoto(list_comprehension, 116) ->
 56;
yeccgoto(list_comprehension, 130) ->
 56;
yeccgoto(list_comprehension, 134) ->
 56;
yeccgoto(list_comprehension, 146) ->
 56;
yeccgoto(list_comprehension, 151) ->
 56;
yeccgoto(list_comprehension, 160) ->
 56;
yeccgoto(list_comprehension, 167) ->
 56;
yeccgoto(list_comprehension, 172) ->
 56;
yeccgoto(list_comprehension, 185) ->
 56;
yeccgoto(list_comprehension, 187) ->
 56;
yeccgoto(list_comprehension, 189) ->
 56;
yeccgoto(list_comprehension, 191) ->
 56;
yeccgoto(list_comprehension, 192) ->
 56;
yeccgoto(list_comprehension, 195) ->
 56;
yeccgoto(list_comprehension, 199) ->
 56;
yeccgoto(list_comprehension, 206) ->
 56;
yeccgoto(list_comprehension, 209) ->
 56;
yeccgoto(list_comprehension, 219) ->
 56;
yeccgoto(list_comprehension, 221) ->
 56;
yeccgoto(list_comprehension, 235) ->
 56;
yeccgoto(list_comprehension, 243) ->
 56;
yeccgoto(list_comprehension, 261) ->
 56;
yeccgoto(list_op, 40) ->
 172;
yeccgoto(mult_op, 41) ->
 160;
yeccgoto(mult_op, 176) ->
 160;
yeccgoto(opt_bit_size_expr, 217) ->
 222;
yeccgoto(opt_bit_type_list, 222) ->
 224;
yeccgoto(prefix_op, 14) ->
 58;
yeccgoto(prefix_op, 18) ->
 58;
yeccgoto(prefix_op, 20) ->
 58;
yeccgoto(prefix_op, 23) ->
 219;
yeccgoto(prefix_op, 24) ->
 58;
yeccgoto(prefix_op, 27) ->
 58;
yeccgoto(prefix_op, 30) ->
 58;
yeccgoto(prefix_op, 32) ->
 58;
yeccgoto(prefix_op, 52) ->
 58;
yeccgoto(prefix_op, 61) ->
 58;
yeccgoto(prefix_op, 66) ->
 58;
yeccgoto(prefix_op, 70) ->
 58;
yeccgoto(prefix_op, 75) ->
 58;
yeccgoto(prefix_op, 81) ->
 58;
yeccgoto(prefix_op, 85) ->
 58;
yeccgoto(prefix_op, 88) ->
 58;
yeccgoto(prefix_op, 90) ->
 58;
yeccgoto(prefix_op, 98) ->
 58;
yeccgoto(prefix_op, 102) ->
 58;
yeccgoto(prefix_op, 107) ->
 58;
yeccgoto(prefix_op, 109) ->
 58;
yeccgoto(prefix_op, 116) ->
 58;
yeccgoto(prefix_op, 130) ->
 58;
yeccgoto(prefix_op, 146) ->
 58;
yeccgoto(prefix_op, 151) ->
 58;
yeccgoto(prefix_op, 160) ->
 58;
yeccgoto(prefix_op, 167) ->
 58;
yeccgoto(prefix_op, 172) ->
 58;
yeccgoto(prefix_op, 185) ->
 58;
yeccgoto(prefix_op, 187) ->
 58;
yeccgoto(prefix_op, 189) ->
 58;
yeccgoto(prefix_op, 191) ->
 58;
yeccgoto(prefix_op, 192) ->
 58;
yeccgoto(prefix_op, 195) ->
 58;
yeccgoto(prefix_op, 199) ->
 58;
yeccgoto(prefix_op, 206) ->
 58;
yeccgoto(prefix_op, 209) ->
 58;
yeccgoto(prefix_op, 235) ->
 219;
yeccgoto(prefix_op, 243) ->
 58;
yeccgoto(prefix_op, 261) ->
 58;
yeccgoto(query_expr, 14) ->
 60;
yeccgoto(query_expr, 18) ->
 60;
yeccgoto(query_expr, 20) ->
 60;
yeccgoto(query_expr, 23) ->
 60;
yeccgoto(query_expr, 24) ->
 60;
yeccgoto(query_expr, 27) ->
 60;
yeccgoto(query_expr, 30) ->
 60;
yeccgoto(query_expr, 32) ->
 60;
yeccgoto(query_expr, 52) ->
 60;
yeccgoto(query_expr, 58) ->
 60;
yeccgoto(query_expr, 61) ->
 60;
yeccgoto(query_expr, 66) ->
 60;
yeccgoto(query_expr, 70) ->
 60;
yeccgoto(query_expr, 75) ->
 60;
yeccgoto(query_expr, 81) ->
 60;
yeccgoto(query_expr, 85) ->
 60;
yeccgoto(query_expr, 88) ->
 60;
yeccgoto(query_expr, 90) ->
 60;
yeccgoto(query_expr, 98) ->
 60;
yeccgoto(query_expr, 102) ->
 60;
yeccgoto(query_expr, 107) ->
 60;
yeccgoto(query_expr, 109) ->
 60;
yeccgoto(query_expr, 116) ->
 60;
yeccgoto(query_expr, 130) ->
 60;
yeccgoto(query_expr, 134) ->
 60;
yeccgoto(query_expr, 146) ->
 60;
yeccgoto(query_expr, 151) ->
 60;
yeccgoto(query_expr, 160) ->
 60;
yeccgoto(query_expr, 167) ->
 60;
yeccgoto(query_expr, 172) ->
 60;
yeccgoto(query_expr, 185) ->
 60;
yeccgoto(query_expr, 187) ->
 60;
yeccgoto(query_expr, 189) ->
 60;
yeccgoto(query_expr, 191) ->
 60;
yeccgoto(query_expr, 192) ->
 60;
yeccgoto(query_expr, 195) ->
 60;
yeccgoto(query_expr, 199) ->
 60;
yeccgoto(query_expr, 206) ->
 60;
yeccgoto(query_expr, 209) ->
 60;
yeccgoto(query_expr, 219) ->
 60;
yeccgoto(query_expr, 221) ->
 60;
yeccgoto(query_expr, 235) ->
 60;
yeccgoto(query_expr, 243) ->
 60;
yeccgoto(query_expr, 261) ->
 60;
yeccgoto(receive_expr, 14) ->
 62;
yeccgoto(receive_expr, 18) ->
 62;
yeccgoto(receive_expr, 20) ->
 62;
yeccgoto(receive_expr, 23) ->
 62;
yeccgoto(receive_expr, 24) ->
 62;
yeccgoto(receive_expr, 27) ->
 62;
yeccgoto(receive_expr, 30) ->
 62;
yeccgoto(receive_expr, 32) ->
 62;
yeccgoto(receive_expr, 52) ->
 62;
yeccgoto(receive_expr, 58) ->
 62;
yeccgoto(receive_expr, 61) ->
 62;
yeccgoto(receive_expr, 66) ->
 62;
yeccgoto(receive_expr, 70) ->
 62;
yeccgoto(receive_expr, 75) ->
 62;
yeccgoto(receive_expr, 81) ->
 62;
yeccgoto(receive_expr, 85) ->
 62;
yeccgoto(receive_expr, 88) ->
 62;
yeccgoto(receive_expr, 90) ->
 62;
yeccgoto(receive_expr, 98) ->
 62;
yeccgoto(receive_expr, 102) ->
 62;
yeccgoto(receive_expr, 107) ->
 62;
yeccgoto(receive_expr, 109) ->
 62;
yeccgoto(receive_expr, 116) ->
 62;
yeccgoto(receive_expr, 130) ->
 62;
yeccgoto(receive_expr, 134) ->
 62;
yeccgoto(receive_expr, 146) ->
 62;
yeccgoto(receive_expr, 151) ->
 62;
yeccgoto(receive_expr, 160) ->
 62;
yeccgoto(receive_expr, 167) ->
 62;
yeccgoto(receive_expr, 172) ->
 62;
yeccgoto(receive_expr, 185) ->
 62;
yeccgoto(receive_expr, 187) ->
 62;
yeccgoto(receive_expr, 189) ->
 62;
yeccgoto(receive_expr, 191) ->
 62;
yeccgoto(receive_expr, 192) ->
 62;
yeccgoto(receive_expr, 195) ->
 62;
yeccgoto(receive_expr, 199) ->
 62;
yeccgoto(receive_expr, 206) ->
 62;
yeccgoto(receive_expr, 209) ->
 62;
yeccgoto(receive_expr, 219) ->
 62;
yeccgoto(receive_expr, 221) ->
 62;
yeccgoto(receive_expr, 235) ->
 62;
yeccgoto(receive_expr, 243) ->
 62;
yeccgoto(receive_expr, 261) ->
 62;
yeccgoto(record_expr, 14) ->
 63;
yeccgoto(record_expr, 18) ->
 63;
yeccgoto(record_expr, 20) ->
 63;
yeccgoto(record_expr, 24) ->
 63;
yeccgoto(record_expr, 27) ->
 63;
yeccgoto(record_expr, 30) ->
 63;
yeccgoto(record_expr, 32) ->
 63;
yeccgoto(record_expr, 52) ->
 63;
yeccgoto(record_expr, 58) ->
 63;
yeccgoto(record_expr, 61) ->
 63;
yeccgoto(record_expr, 66) ->
 63;
yeccgoto(record_expr, 70) ->
 63;
yeccgoto(record_expr, 75) ->
 63;
yeccgoto(record_expr, 81) ->
 63;
yeccgoto(record_expr, 85) ->
 63;
yeccgoto(record_expr, 88) ->
 63;
yeccgoto(record_expr, 90) ->
 63;
yeccgoto(record_expr, 98) ->
 63;
yeccgoto(record_expr, 102) ->
 63;
yeccgoto(record_expr, 107) ->
 63;
yeccgoto(record_expr, 109) ->
 63;
yeccgoto(record_expr, 116) ->
 63;
yeccgoto(record_expr, 130) ->
 63;
yeccgoto(record_expr, 146) ->
 63;
yeccgoto(record_expr, 151) ->
 63;
yeccgoto(record_expr, 160) ->
 63;
yeccgoto(record_expr, 167) ->
 63;
yeccgoto(record_expr, 172) ->
 63;
yeccgoto(record_expr, 185) ->
 63;
yeccgoto(record_expr, 187) ->
 63;
yeccgoto(record_expr, 189) ->
 63;
yeccgoto(record_expr, 191) ->
 63;
yeccgoto(record_expr, 192) ->
 63;
yeccgoto(record_expr, 195) ->
 63;
yeccgoto(record_expr, 199) ->
 63;
yeccgoto(record_expr, 206) ->
 63;
yeccgoto(record_expr, 209) ->
 63;
yeccgoto(record_expr, 243) ->
 63;
yeccgoto(record_expr, 261) ->
 63;
yeccgoto(record_field, 140) ->
 142;
yeccgoto(record_field, 149) ->
 142;
yeccgoto(record_fields, 140) ->
 143;
yeccgoto(record_fields, 149) ->
 150;
yeccgoto(record_tuple, 137) ->
 139;
yeccgoto(record_tuple, 239) ->
 241;
yeccgoto(rule, 0) ->
 8;
yeccgoto(rule_body, 17) ->
 244;
yeccgoto(rule_body, 259) ->
 244;
yeccgoto(rule_clause, 0) ->
 9;
yeccgoto(rule_clause, 11) ->
 9;
yeccgoto(rule_clauses, 0) ->
 10;
yeccgoto(rule_clauses, 11) ->
 13;
yeccgoto(strings, 14) ->
 65;
yeccgoto(strings, 18) ->
 65;
yeccgoto(strings, 20) ->
 65;
yeccgoto(strings, 23) ->
 65;
yeccgoto(strings, 24) ->
 65;
yeccgoto(strings, 27) ->
 65;
yeccgoto(strings, 30) ->
 65;
yeccgoto(strings, 32) ->
 65;
yeccgoto(strings, 52) ->
 65;
yeccgoto(strings, 58) ->
 65;
yeccgoto(strings, 61) ->
 65;
yeccgoto(strings, 64) ->
 87;
yeccgoto(strings, 66) ->
 65;
yeccgoto(strings, 70) ->
 65;
yeccgoto(strings, 75) ->
 65;
yeccgoto(strings, 81) ->
 65;
yeccgoto(strings, 85) ->
 65;
yeccgoto(strings, 88) ->
 65;
yeccgoto(strings, 90) ->
 65;
yeccgoto(strings, 98) ->
 65;
yeccgoto(strings, 102) ->
 65;
yeccgoto(strings, 107) ->
 65;
yeccgoto(strings, 109) ->
 65;
yeccgoto(strings, 116) ->
 65;
yeccgoto(strings, 130) ->
 65;
yeccgoto(strings, 134) ->
 65;
yeccgoto(strings, 146) ->
 65;
yeccgoto(strings, 151) ->
 65;
yeccgoto(strings, 160) ->
 65;
yeccgoto(strings, 167) ->
 65;
yeccgoto(strings, 172) ->
 65;
yeccgoto(strings, 185) ->
 65;
yeccgoto(strings, 187) ->
 65;
yeccgoto(strings, 189) ->
 65;
yeccgoto(strings, 191) ->
 65;
yeccgoto(strings, 192) ->
 65;
yeccgoto(strings, 195) ->
 65;
yeccgoto(strings, 199) ->
 65;
yeccgoto(strings, 206) ->
 65;
yeccgoto(strings, 209) ->
 65;
yeccgoto(strings, 219) ->
 65;
yeccgoto(strings, 221) ->
 65;
yeccgoto(strings, 235) ->
 65;
yeccgoto(strings, 243) ->
 65;
yeccgoto(strings, 261) ->
 65;
yeccgoto(tail, 205) ->
 208;
yeccgoto(tail, 212) ->
 213;
yeccgoto(try_expr, 14) ->
 67;
yeccgoto(try_expr, 18) ->
 67;
yeccgoto(try_expr, 20) ->
 67;
yeccgoto(try_expr, 23) ->
 67;
yeccgoto(try_expr, 24) ->
 67;
yeccgoto(try_expr, 27) ->
 67;
yeccgoto(try_expr, 30) ->
 67;
yeccgoto(try_expr, 32) ->
 67;
yeccgoto(try_expr, 52) ->
 67;
yeccgoto(try_expr, 58) ->
 67;
yeccgoto(try_expr, 61) ->
 67;
yeccgoto(try_expr, 66) ->
 67;
yeccgoto(try_expr, 70) ->
 67;
yeccgoto(try_expr, 75) ->
 67;
yeccgoto(try_expr, 81) ->
 67;
yeccgoto(try_expr, 85) ->
 67;
yeccgoto(try_expr, 88) ->
 67;
yeccgoto(try_expr, 90) ->
 67;
yeccgoto(try_expr, 98) ->
 67;
yeccgoto(try_expr, 102) ->
 67;
yeccgoto(try_expr, 107) ->
 67;
yeccgoto(try_expr, 109) ->
 67;
yeccgoto(try_expr, 116) ->
 67;
yeccgoto(try_expr, 130) ->
 67;
yeccgoto(try_expr, 134) ->
 67;
yeccgoto(try_expr, 146) ->
 67;
yeccgoto(try_expr, 151) ->
 67;
yeccgoto(try_expr, 160) ->
 67;
yeccgoto(try_expr, 167) ->
 67;
yeccgoto(try_expr, 172) ->
 67;
yeccgoto(try_expr, 185) ->
 67;
yeccgoto(try_expr, 187) ->
 67;
yeccgoto(try_expr, 189) ->
 67;
yeccgoto(try_expr, 191) ->
 67;
yeccgoto(try_expr, 192) ->
 67;
yeccgoto(try_expr, 195) ->
 67;
yeccgoto(try_expr, 199) ->
 67;
yeccgoto(try_expr, 206) ->
 67;
yeccgoto(try_expr, 209) ->
 67;
yeccgoto(try_expr, 219) ->
 67;
yeccgoto(try_expr, 221) ->
 67;
yeccgoto(try_expr, 235) ->
 67;
yeccgoto(try_expr, 243) ->
 67;
yeccgoto(try_expr, 261) ->
 67;
yeccgoto(tuple, 14) ->
 68;
yeccgoto(tuple, 18) ->
 68;
yeccgoto(tuple, 20) ->
 68;
yeccgoto(tuple, 23) ->
 68;
yeccgoto(tuple, 24) ->
 68;
yeccgoto(tuple, 27) ->
 68;
yeccgoto(tuple, 30) ->
 68;
yeccgoto(tuple, 32) ->
 68;
yeccgoto(tuple, 52) ->
 68;
yeccgoto(tuple, 58) ->
 68;
yeccgoto(tuple, 61) ->
 68;
yeccgoto(tuple, 66) ->
 68;
yeccgoto(tuple, 70) ->
 68;
yeccgoto(tuple, 75) ->
 68;
yeccgoto(tuple, 81) ->
 68;
yeccgoto(tuple, 85) ->
 68;
yeccgoto(tuple, 88) ->
 68;
yeccgoto(tuple, 90) ->
 68;
yeccgoto(tuple, 98) ->
 68;
yeccgoto(tuple, 102) ->
 68;
yeccgoto(tuple, 107) ->
 68;
yeccgoto(tuple, 109) ->
 68;
yeccgoto(tuple, 116) ->
 68;
yeccgoto(tuple, 130) ->
 68;
yeccgoto(tuple, 134) ->
 68;
yeccgoto(tuple, 146) ->
 68;
yeccgoto(tuple, 151) ->
 68;
yeccgoto(tuple, 160) ->
 68;
yeccgoto(tuple, 167) ->
 68;
yeccgoto(tuple, 172) ->
 68;
yeccgoto(tuple, 185) ->
 68;
yeccgoto(tuple, 187) ->
 68;
yeccgoto(tuple, 189) ->
 68;
yeccgoto(tuple, 191) ->
 68;
yeccgoto(tuple, 192) ->
 68;
yeccgoto(tuple, 195) ->
 68;
yeccgoto(tuple, 199) ->
 68;
yeccgoto(tuple, 206) ->
 68;
yeccgoto(tuple, 209) ->
 68;
yeccgoto(tuple, 219) ->
 68;
yeccgoto(tuple, 221) ->
 68;
yeccgoto(tuple, 235) ->
 68;
yeccgoto(tuple, 243) ->
 68;
yeccgoto(tuple, 261) ->
 68;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


